--- The functions to handle researching traits. Since only storing the timestamp
--  of when research on a trait is completed only stores timestamps on startup
--  and then registers for when a new research is started or canceled to update
--  the timestamp.
--  @module TONResearch
--  @author Tirtow

--- Saves the time that the item that research was started on will be completed.
--  @param trade_skill      the trade line the research is on
--  @param line_index       the index of the item in the trade
--  @param trait_index      the index of the trait being researched
function TON:ResearchStarted(trade_skill, line_index, trait_index)
  local var_name = tostring(trade_skill) .. "_" .. tostring(line_index) ..
                   "_" .. tostring(trait_index)
  local _, time_remaining = GetSmithingResearchLineTraitTimes(trade_skill,
      line_index, trait_index)
  local new_time = GetTimeStamp() + time_remaining
  self.saved_variables.toons[self.toon_id][var_name] = new_time
  self.saved_variables.toons[self.toon_id]["max_" .. tostring(trade_skill)] =
      GetMaxSimultaneousSmithingResearch(trade_skill)

  self:PrintDebug("set completion time for " .. tostring(trade_skill) ..
                  "_" .. tostring(line_index) .. "_" .. tostring(trait_index) ..
                  " to " .. tostring(new_time))
  self:RequestPrioritySave()
  self:NotifyOfUpdate("research completion time")
end

--- Sets the time a research item will be completed back to UNKNOWN_TRAIT.
--  @param trade_skill      the trade line the research is on
--  @param line_index       the index of the item in the trade
--  @param trait_index      the index of the trait that was canceled
--  @param unknown_trait    the timestamp to represent an unknown trait
function TON:ResearchCanceled(trade_skill, line_index, trait_index,
                              unknown_trait)
  local var_name = tostring(trade_skill) .. "_" .. tostring(line_index) ..
                   "_" .. tostring(trait_index)
  self.saved_variables.toons[self.toon_id][var_name] = unknown_trait

  self:PrintDebug("canceled item " .. tostring(trade_skill) .. "_" ..
                  tostring(line_index) .. "_" .. tostring(trait_index))
  self:RequestPrioritySave()
  self:NotifyOfUpdate("canceled research item")
end

--- Saves whether or not each trait has been researched.
--  If the trait has already been researched saves DEFAULT_TIME, if the trait is
--  being researched saves the timestamp of when it will be completed, if the
--  trait is unknown and not being researched saves UNKNOWN_TRAIT.
--  @param trade_skill    the TradeskillType
--  @param default_time   the default timestamp
--  @param unknown_trait  the timestamp to represent an unknown trait
function TON:SaveAllResearchedTraits(trade_skill, default_time, unknown_trait)
  self:PrintDebug("saving researched traits for " .. tostring(trade_skill))
  
  self.saved_variables.toons[self.toon_id]["max_" .. tostring(trade_skill)] =
      GetMaxSimultaneousSmithingResearch(trade_skill)

  for line_index = 1, GetNumSmithingResearchLines(trade_skill) do
    local _, _, num_traits = GetSmithingResearchLineInfo(trade_skill,
        line_index)
    for trait_index = 1, num_traits do
      local _, _, known = GetSmithingResearchLineTraitInfo(trade_skill,
          line_index, trait_index)
      if known then
        self:CompletedResearch(trade_skill, line_index, trait_index,
                               default_time)
      else
        self:UpdateResearchTime(trade_skill, line_index, trait_index,
                                unknown_trait)
      end
    end
  end
end

--- Updates the time remaining on the research of a trait
--  @param trade_skill    the TradeskillType
--  @param line_index     the index of the item in the trade
--  @param trait_index    the index of the trait that was completed
--  @param default_time   the timestamp for an unknown trait
function TON:UpdateResearchTime(trade_skill, line_index, trait_index,
                                unknown_trait)
  local var_name = tostring(trade_skill) .. "_" .. tostring(line_index) ..
                   "_" .. tostring(trait_index)
  local _, time_remaining = GetSmithingResearchLineTraitTimes(trade_skill,
                                                              line_index,
                                                              trait_index)
  if time_remaining ~= nil then
    self.saved_variables.toons[self.toon_id][var_name] = GetTimeStamp() +
      time_remaining
  else
    self.saved_variables.toons[self.toon_id][var_name] = unknown_trait
  end
end

--- Marks the given trait as known
--  @param trade_skill    the TradeskillType
--  @param line_index     the index of the item in the trade
--  @param trait_index    the index of the trait that was completed
--  @param default_time   the default timestamp
function TON:CompletedResearch(trade_skill, line_index, trait_index,
                               default_time)
  local var_name = tostring(trade_skill) .. "_" .. tostring(line_index) ..
                   "_" .. tostring(trait_index)
  self.saved_variables.toons[self.toon_id][var_name] = default_time
end

--- Updates the trait times for every trait for the currect character.
--  @param default_time   the default timestamp
--  @param unknown_trait  the timestamp to represent an unknown trait
function TON:SaveAllResearch(default_time, unknown_trait)
  self:SaveAllResearchedTraits(CRAFTING_TYPE_BLACKSMITHING,
                              default_time, unknown_trait)
  self:SaveAllResearchedTraits(CRAFTING_TYPE_CLOTHIER,
                              default_time, unknown_trait)
  self:SaveAllResearchedTraits(CRAFTING_TYPE_WOODWORKING,
                              default_time, unknown_trait)
  self:SaveAllResearchedTraits(CRAFTING_TYPE_JEWELRYCRAFTING,
                              default_time, unknown_trait)
end
