--- The functions to determine if a trial was completed and to update the
--  timestamp for the completed trial if one was found. Since there is no API
--  method to get the cooldown for a trial coffer, the addon must be running
--  when the coffer is received and will scan the character's inventory to find
--  the coffer matching the trial quest that was turned in.
--  @module TONTrials
--  @author Tirtow

--- Number of seconds between trial coffers
local SECS_FOR_TRIAL_COOLDOWN = 7 * 24 * 60 * 60

--- Constants to identify trials
--  @field HEL_RA_CITADEL           index of hrc in TRIALS
--  @field AETHERIAN_ARCHIVE        index of aa  in TRIALS
--  @field SANCTUM_OPHIDIA          index of so  in TRIALS
--  @field MAW_OF_LORKHAJ           index of maw in TRIALS
--  @field HALLS_OF_FABRICATION     index of hof in TRIALS
--  @field ASYLUM_SANCTORIUM        index of as  in TRIALS
--  @field CLOUDREST                index of cloudreset in TRIALS
--  @field TRIALS                   table of trial info
local HEL_RA_CITADEL       = 0
local AETHERIAN_ARCHIVE    = 1
local SANCTUM_OPHIDIA      = 2
local MAW_OF_LORKHAJ       = 3
local HALLS_OF_FABRICATION = 4
local ASYLUM_SANCTORIUM    = 5
local CLOUDREST            = 6
local TRIALS = {}
TRIALS[HEL_RA_CITADEL] = {
  var_name    = "hel_ra_citadel",
  name        = "Hel Ra Citadel",
  coffer_name = "warrior's dulled coffer"
}
TRIALS[AETHERIAN_ARCHIVE] = {
  var_name    = "aetherian_archive",
  name        = "Aetherian Archive",
  coffer_name = "mage's ignorant coffer"
}
TRIALS[SANCTUM_OPHIDIA] = {
  var_name    = "sanctum_ophidia",
  name        = "Sanctum Ophidia",
  coffer_name = "serpent's languid coffer"
}
TRIALS[MAW_OF_LORKHAJ] = {
  var_name    = "maw_of_lorkhaj",
  name        = "Maw of Lorkhaj",
  coffer_name = "dro-m'athra's burnished coffer"
}
TRIALS[HALLS_OF_FABRICATION] = {
  var_name    = "halls_of_fabrication",
  name        = "Halls of Fabrication",
  coffer_name = "fabricant's burnished coffer"
}
TRIALS[ASYLUM_SANCTORIUM] = {
  var_name    = "asylum_sanctorium",
  name        = "Asylum Sanctorium",
  coffer_name = "saint's beatified coffer"
}
TRIALS[CLOUDREST] = {
  var_name    = "cloudrest",
  name        = "Cloudrest",
  coffer_name = "welkynar's grounded coffer"
}

--- Gets the index for the trial given the name of the quest for the trial
--  @param quest_name     the name of the quest
--  @return the index of the trial in TRIALS or nil if not a trial
function TON:GetTrialByQuest(quest_name)
  quest_name = string.lower(quest_name)
  self:PrintDebug("quest_name: " .. quest_name)

  if quest_name == "assaulting the citadel" then
    self:PrintDebug("completed HRC")
    return HEL_RA_CITADEL
  elseif quest_name == "the mage's tower" then
    self:PrintDebug("completed AA")
    return AETHERIAN_ARCHIVE
  elseif quest_name == "the oldest ghost" then
    self:PrintDebug("completed SO")
    return SANCTUM_OPHIDIA
  elseif quest_name == "into the maw" then
    self:PrintDebug("completed MAW")
    return MAW_OF_LORKHAJ
  elseif quest_name == "forging the future" then
    self:PrintDebug("completed HOF")
    return HALLS_OF_FABRICATION
  elseif quest_name == "saints' mercy" then
    self:PrintDebug("completed AS")
    return ASYLUM_SANCTORIUM
  elseif quest_name == "woe of the welkynars" then
    self:PrintDebug("completed cloudrest")
    return CLOUDREST
  else
    self:PrintDebug("trial not found")
    return nil
  end
end

--- Update the trial coffer timer for the current character.
--  @param trial_index    the index in TRIALS of the trial that was just
--                        completed
function TON:SaveCofferTime(trial_index)
  local trial = TRIALS[trial_index]
  local next_time = GetTimeStamp() + SECS_FOR_TRIAL_COOLDOWN
  self.saved_variables.toons[self.toon_id][trial.var_name] = next_time
  self:RequestPrioritySave()
  self:NotifyOfUpdate("coffer timer for " .. trial.name)

  self:PrintDebug("set coffer time for " .. trial.name .. " to " ..
                  tostring(next_time))
end

--- Checks if the user just received a trial coffer by looping through their
--  inventory looking for the coffer that corresponds to the completed trial
--  quest.
--  @param quest_name    the name of the completed quest
function TON:CheckIfGotCoffer(quest_name)
  local trial = self:GetTrialByQuest(quest_name)

  if trial then
    local coffer_name = TRIALS[trial].coffer_name
    local inv_size = GetBagSize(BAG_BACKPACK)

    self:PrintDebug("trial is "       .. tostring(trial))
    self:PrintDebug("coffer_name is " .. coffer_name)

    for i = 1, inv_size do
      if string.lower(GetItemName(BAG_BACKPACK, i)) == coffer_name then
        local timer = self.saved_variables.toons[self.toon_id][TRIALS[trial].var_name]
        if timer == nil or timer <= GetTimeStamp() then
          self:SaveCofferTime(trial)
        end
        break
      end
    end
  else
    self:PrintDebug("failed to find trial")
  end
end
