--- The addon to handle communicating between ESO and the client
--  Saves any information needing to be sent to the server to the saved
--  variables for the client to push to the server
--  @module TimersOfNirn_v1.4.0
--  @author Tirtow

--- Default time for time stamps
local DEFAULT_TIME            = 1502863200
--- Time to wait before updating random time after LFG
local UPDATE_DELAY            = 10 * 1000
--- Number of seconds between LFG daily rewards
local LFG_COOLDOWN            = (20 * 60 * 60)
--- Time to use if trait unknown, 1/1/3000, will need to change if Molag Bal
--  has not taken over Nirn by then
local UNKNOWN_TRAIT           = 32503680000

--- TON table
--  @field name              addon name
--  @field version           addon version
--  @field variable_version  saved variables version
--  @field toon_id           the ID of the current character
--  @field drt_visible       whether or not the DRT UI is visible
--  @field horse_visible     whether or not the horse UI is visible
--  @field default           default saved variables
TON = {}
TON.name             = "TimersOfNirn"
TON.version          = "1.4.1"
TON.variable_version = 1
TON.toon_id          = GetCurrentCharacterId()
TON.drt_visible      = false
TON.horse_visible    = false
TON.default = {
  debug         = false,          -- Whether or not to show debug output
  show_messages = true,           -- Whether or not to show update messages
  toons = {
    [TON.toon_id] = {
      alive           = false,    -- Used to track if character was deleted
      addon_id        = -1,       -- The id       of this character
      name            = "none",   -- The name     of this character
      race            = "none",   -- The race     of this character
      alliance        = "none",   -- The alliance of this character
      level           = 0,        -- The level    of this character
      champion_points = 0,        -- The account champion points
      toon_class      = "none",   -- The class of this character
      seconds_played  = 0,              -- Number of seconds played
      horse                = DEFAULT_TIME,    -- Next time for mount upgrade
      random               = DEFAULT_TIME,    -- Next time for premium LFG
      updated_at           = DEFAULT_TIME,    -- Time stamp of last time updated
      aetherian_archive    = DEFAULT_TIME,    -- coffer timer for AA
      sanctum_ophidia      = DEFAULT_TIME,    -- coffer timer for SO
      hel_ra_citadel       = DEFAULT_TIME,    -- coffer timer for HRC
      maw_of_lorkhaj       = DEFAULT_TIME,    -- coffer timer for MAW
      halls_of_fabrication = DEFAULT_TIME,    -- coffer timer for HOF
      asylum_sanctorium    = DEFAULT_TIME,    -- coffer timer for AS
      cloudrest            = DEFAULT_TIME,    -- coffer timer for cloudrest
      blacksmithing_hireling = DEFAULT_TIME,    -- hireling timer for blacksmith
      clothing_hireling      = DEFAULT_TIME,    -- hireling timer for clothing
      woodworking_hireling   = DEFAULT_TIME,    -- hireling timer for woodwork
      provisioning_hireling  = DEFAULT_TIME,    -- hireling timer for provisioning
      enchanting_hireling    = DEFAULT_TIME,    -- hireling timer for enchanting
      jewelry_hireling       = DEFAULT_TIME,    -- hireling timer for jewelry
    }
  }
}

--- Requests for the addon to be prioritized for saving variables
function TON:RequestPrioritySave()
  self:PrintDebug("requesting priority save")
  self.saved_variables.toons[self.toon_id].updated_at = GetTimeStamp()
  -- RequestAddOnSavedVariablesPrioritySave(self.name)
end

--- Notifies the user that one of the things saved by the addon was updated and
--  that the client will push the updates once the UI is reloaded or user logs
--  out.
--  @param part     the part that was updated
function TON:NotifyOfUpdate(part)
  if self.saved_variables.show_messages then
    d("[TON]: updated |cFFFFFF" .. part .. "|r. Changes will be pushed to " ..
      "the server on /reloadui or log out")
  end
end

--- Saves the next time a daily random dungeon is able to be ran for the full
--  premium reward.
function TON:SaveRandomTime()
  local is_eligible = IsEligibleForDailyActivityReward()

  self:PrintDebug("is_eligible = " .. tostring(is_eligible))
  self:PrintDebug("was_eligible = " .. tostring(self.was_eligible))

  if not is_eligible and self.was_eligible then
    self:UpdateRandomTimestamp()
  end
end

--- Updates the timestamp for the daily LFG random
function TON:UpdateRandomTimestamp()
    self.saved_variables.toons[self.toon_id].random = GetTimeStamp() +
      LFG_COOLDOWN
    self:RequestPrioritySave()
    self:NotifyOfUpdate("daily random time")
    self.was_eligible = false
    self:PrintDebug("set random to " ..
                    tostring(self.saved_variables.toons[self.toon_id].random))
end

--- Saves the next time the character's horse can be upgraded
--  If the character has fully completed horse training, sets the horse
--  timestamp to DEFAULT_TIME
function TON:SaveHorseTime()
  local inv, max_inv, stam, max_stam, speed, max_speed = GetRidingStats()
  if inv == max_inv and stam == max_stam and speed == max_speed then
    self.saved_variables.toons[self.toon_id].horse = DEFAULT_TIME
    self:PrintDebug("character has completed horse training")
  else
    local horse_time, _ = GetTimeUntilCanBeTrained()
    self.saved_variables.toons[self.toon_id].horse = GetTimeStamp() +
        (horse_time / 1000)

    self:PrintDebug("set horse to " ..
        tostring(self.saved_variables.toons[self.toon_id].horse))
  end
end

--- Saves character information on user logout.
--  Updates level/champion points, seconds played, and the daily random.
--  Daily random is saved on logout in case the event to pick up when a LFG is
--  completed does not store the correct timestamp.
function TON:SaveLogoutInfo()
  self.saved_variables.toons[self.toon_id].updated_at      = GetTimeStamp()
  self.saved_variables.toons[self.toon_id].seconds_played  = GetSecondsPlayed()
  self.saved_variables.toons[self.toon_id].champion_points = GetUnitChampionPoints(
      "player")
  self.saved_variables.toons[self.toon_id].title = GetTitle(
    GetCurrentTitleIndex())
    
  for i = 1, GetNumCharacters() do
    local _, _, level, _, _, _, id, _ = GetCharacterInfo(i)
    if id == self.toon_id then
      self.saved_variables.toons[self.toon_id].level = level
      break
    end
  end
  self:SaveHorseTime()
end

--- Checks if the user completed a LFG dungeon and saves the time until the next
--  premium reward can be earned
--  @param activity_type     the type of activity
function TON:LFGCompleted(activity_type)
  self:PrintDebug("completed activity " .. tostring(activity_type))
  if activity_type == LFG_ACTIVITY_DUNGEON or
     activity_type == LFG_ACTIVITY_MASTER_DUNGEON then
    self:SaveRandomTime()
  end
end

--- Callback to handle when the player is activated
--  Saves the basic character info, horse timer, and updates researched traits.
--  Sets this character's addon_id if this is the first time the addon has seen
--  this character.
local function ActivatedUpdate(event, initial)
  EVENT_MANAGER:UnregisterForEvent(TON.name, EVENT_PLAYER_ACTIVATED)
  TON:CreateSettingsWindow()
  TON.was_eligible = IsEligibleForDailyActivityReward()
  TON:PrintDebug("was_eligible = " .. tostring(TON.was_eligible))
        
  if initial then
    TON:PrintDebug("handling initial login")
    TON:ClearAlive()
    TON:SaveBasicInfo()
    TON:SaveHorseTime()
    TON:SaveAllResearch(DEFAULT_TIME, UNKNOWN_TRAIT)
    TON:RemoveDeadCharacters()
    TON:StartupMessage()
  end
end

--- Callback to handle when the user is deactivated
--  Saves the basic character information and updates what timers can be
--  updated
local function DeactivatedUpdate()
  EVENT_MANAGER:UnregisterForEvent(TON.name, EVENT_PLAYER_DEACTIVATED)
  TON:SaveLogoutInfo()
end

--- Callback to handle when a quest is completed to check if the user just
--  completed a trial quest and received a coffer
--  @param event_code    the event
--  @param quest_name    the name of the quest that was completed
local function QuestCompleted(event_code, quest_name, _, _, _, _, _, _)
  TON:PrintDebug("quest_name is " .. quest_name)
  TON:CheckIfGotCoffer(quest_name)
end

--- Callback to handle when a LFG instance is completed.
--  Waits UPDATE_DELAY seconds then saves the next daily random timestamp
local function LFGCompleted()
  TON:PrintDebug("completed a LFG instance")

  zo_callLater(
      function() 
        TON:LFGCompleted(GetActivityType(GetCurrentLFGActivityId()))
      end,
      UPDATE_DELAY
  )
end

--- Callback to handle when the character's horse attributes are upgraded
--  @param event_code   the event code
local function HorseUpdated(event_code)
  TON:PrintDebug("horse was updated")
  TON:SaveHorseTime()
  TON:RequestPrioritySave()
  TON:NotifyOfUpdate("horse feed timer")
end

--- Callback to handle when the character starts researching a new item.
--  @param event_code       the event code
--  @param trade_skill      the trade line the research is on
--  @param line_index       the index of the item in the trade
--  @param trait_index      the index of the trait being researched
local function ResearchStarted(event_code, trade_skill, line_index, trait_index)
  TON:PrintDebug("research was started")
  TON:ResearchStarted(trade_skill, line_index, trait_index)
end

--- Callback to handle when the character cancels researching a new item.
--  @param event_code       the event code
--  @param trade_skill      the trade line the research is on
--  @param line_index       the index of the item in the trade
--  @param trait_index      the index of the trait that was canceled
local function ResearchCanceled(event_code, trade_skill, line_index,
                                trait_index)
  TON:PrintDebug("research was canceled")
  TON:ResearchCanceled(trade_skill, line_index, trait_index, UNKNOWN_TRAIT)
end

--- Callback to handle when the research times for a character is updated
--  @param event_code       the event code
local function ResearchUpdated(event_code)
  TON:PrintDebug("research time updated")
  TON:SaveAllResearch(default_time, unknown_trait)
  TON:RequestPrioritySave()
  TON:NotifyOfUpdate("research times")
end

--- Callback to handle when the character completes researching a new item.
--  @param event_code       the event code
--  @param trade_skill      the trade line the research is on
--  @param line_index       the index of the item in the trade
--  @param trait_index      the index of the trait that was completed
local function ResearchCompleted(event_code, trade_skill, line_index,
                                 trait_index)
  TON:PrintDebug("completed research on " .. tostring(trade_skill) ..
                 "_" .. tostring(line_index) .. "_" ..tostring(trait_index))
  TON:CompletedResearch(trade_skill, line_index, trait_index, DEFAULT_TIME)
  TON:RequestPrioritySave()
  TON:NotifyOfUpdate("research times")
end

--- Callback to handle checking if the character received a plunder skull
--  @param item_name    the name of the item received
local function LootReceived(_, toon_name, item_name, _, _, _, _)
  TON:CheckIfGotPlunderSkull(toon_name, item_name)
end

--- Callback to handle checking if received a hireling mail
--  @param mail_id  the ID of the mail took attachments from
local function MailAttachmentsReceived(_, mail_id)
  TON:PrintDebug("received mail " .. tostring(mail_id))
  TON:CheckForHireling(mail_id)
end

--- Toggles whether or not to print debug output to chat
--  @param set  whether or not to print debug output
function TON:ToggleDebug(set)
  self.saved_variables.debug = set
  d("[TON]: Debug set to " .. tostring(self.saved_variables.debug))
end

--- Registers the slash commands to be used by TON
function TON:RegisterSlashCommands()
  SLASH_COMMANDS["/tondebugenable"]    = function() TON:ToggleDebug(true)  end
  SLASH_COMMANDS["/tondebugdisable"]   = function() TON:ToggleDebug(false) end
  SLASH_COMMANDS["/drt"]               = function() TON:OpenDRTUI()        end
  SLASH_COMMANDS["/hft"]               = function() TON:OpenHorseUI()      end
  SLASH_COMMANDS["/tonupdaterandom"]   = function() TON:UpdateRandomTimestamp() end
  SLASH_COMMANDS["/tonupdateresearch"] = function()
    TON:SaveAllResearch(DEFAULT_TIME, UNKNOWN_TRAIT)
    TON:NotifyOfUpdate("research")
  end
end

--- Registers TON for the needed events
function TON:RegisterForEvents()
  -- Register for when user loads in
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_PLAYER_ACTIVATED,
                                 ActivatedUpdate)

  -- Register for when user is deactivated
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_PLAYER_DEACTIVATED,
                                 DeactivatedUpdate)

  -- Register for when quest is completed
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_QUEST_COMPLETE,
                                 QuestCompleted)

  -- Register for LFG complete
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_ACTIVITY_FINDER_ACTIVITY_COMPLETE,
                                 LFGCompleted)

  -- Register for horse upgrade
  EVENT_MANAGER:RegisterForEvent(TON.name, EVENT_MOUNT_INFO_UPDATED,
                                 HorseUpdated)

  -- Register for research started
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_SMITHING_TRAIT_RESEARCH_STARTED,
                                 ResearchStarted)

  -- Register for research canceled
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_SMITHING_TRAIT_RESEARCH_CANCELED,
                                 ResearchCanceled)

  -- Register for research time updated
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_SMITHING_TRAIT_RESEARCH_TIMES_UPDATED,
                                 ResearchUpdated)
  
  -- Register for when picked up loot  
  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_LOOT_RECEIVED,
                                 LootReceived)

  EVENT_MANAGER:RegisterForEvent(TON.name,
                                 EVENT_MAIL_TAKE_ATTACHED_ITEM_SUCCESS,
                                 MailAttachmentsReceived)
end

--- Initializes the saved variables and slash commands
--  Saved variables are by character
function TON:Initialize()
  EVENT_MANAGER:UnregisterForEvent(TON.name, EVENT_ADD_ON_LOADED)

  -- Get saved variables
  TON.saved_variables = ZO_SavedVars:NewAccountWide("TimersOfNirnVars",
                                                    TON.variable_version,
                                                    nil,
                                                    TON.default)

  TON:RegisterSlashCommands()
  TON:RegisterForEvents()
end

--- Initialize addon
--  @param event    event
--  @param name     addon name
local function OnAddonLoaded(event, name)
  if name == TON.name then
    TON:Initialize()
  end
end

-- Register addon load handler
EVENT_MANAGER:RegisterForEvent(TON.name, EVENT_ADD_ON_LOADED, OnAddonLoaded)
